import asyncio
import io
from concurrent.futures import ProcessPoolExecutor
from itertools import groupby
import numpy as np
import bitstring
from linuxpy.video.device import Device
from Crypto.Hash import BLAKE2b

from loguru import logger

logger.level("INFO")


def split_array_numpy(data):
    data_np = np.frombuffer(data, dtype=np.uint32)
    first = data_np[::2].tobytes()
    second = data_np[1::2].tobytes()
    return first, second


def xor_arrays_numpy(arr1, arr2):
    a = np.frombuffer(arr1, dtype=np.uint32)
    b = np.frombuffer(arr2, dtype=np.uint32)
    return np.bitwise_xor(a, b).tobytes()


def chunks(in_bytes: bytes, take_n: int) -> bytes:
    bytes_buffer = io.BytesIO(in_bytes)
    bytes_send = bytes_buffer.read(take_n)
    while len(bytes_send) == take_n:
        yield bytes_send
        bytes_send = bytes_buffer.read(take_n)


def extract_lsbs(data):
    logger.info("Extract LSBs.")

    buffer = []

    data = [k for k, _ in groupby(data)]
    data = bytes(data)
    data = chunks(data, 8)

    for chunk in data:
        if len(chunk) != 8:
            break

        ba = bitstring.BitArray(8)
        for i, byte in zip(range(len(chunk)), chunk):
            ba[i] = byte & 1

        buffer.append(ba.u)

    return bytes(buffer)


def whiten(data: bytes) -> bytes:
    logger.info("Whitening.")

    buffer = b""
    for chunk in chunks(data, 32):
        if len(chunk) < 32:
            break

        buffer += BLAKE2b.new(data=chunk, digest_bits=256).digest()

    return buffer


def video2_extractor(frame_bytes: bytes):
    logger.debug(f"frames_in {len(frame_bytes)}")
    # chunked_bytes = chunks(frame_bytes, 8)
    # newdata = b""

    first_part, second_part = split_array_numpy(frame_bytes)
    new_data = xor_arrays_numpy(first_part, second_part)
    # for chunk in chunked_bytes:
    #     newdata += bytes(a ^ b for a, b in zip(chunk[:4], chunk[4:]))

    logger.debug(f"new data {len(new_data)}")
    out_bytes = extract_lsbs(new_data)
    out_bytes = whiten(out_bytes)

    logger.info(f"Sample ready: {len(out_bytes)}b.")
    return out_bytes


class FrameTakeWorker:
    def __init__(self, device_id=0, processes=3):
        self.processes = processes
        self.executor_pool = ProcessPoolExecutor(max_workers=3)
        logger.info(self.executor_pool)
        self.video_source = Device.from_id(device_id)
        self.video_source.open()

        self.video_source.set_format(
            1,
            self.video_source.info.frame_sizes[0].width,
            self.video_source.info.frame_sizes[0].height,
            pixel_format="YUYV",
        )

        self.frame_gen = self.video_source.__iter__()

    async def get_frame(self):
        return next(self.frame_gen)

    def __del__(self):
        if self.frame_gen:
            self.frame_gen.close()

    async def frame_extractor(self, frame_bytes: bytes):
        loop = asyncio.get_event_loop()
        logger.info(f"Input bytes: {len(frame_bytes)}")
        output = await loop.run_in_executor(self.executor_pool, video2_extractor, frame_bytes)
        return output

    async def generator_worker(self):
        while True:
            logger.info("get frame")
            frame = await self.get_frame()
            logger.info(f"{frame}")
            data = await self.frame_extractor(bytes(frame))
            logger.info(f"{len(data)}")

    async def run_workers(self):
        tasks = [self.generator_worker() for _ in range(self.processes)]
        await asyncio.gather(*tasks)
        pass


if __name__ == "__main__":
    ftw = FrameTakeWorker(1)
    asyncio.run(ftw.run_workers())
